import logging
from typing import Dict

from .stores import Store, StoreManager

logger = logging.getLogger("gldb")


class GenericLinkedDatabase:

    def __init__(
            self,
            stores: Dict[str, Store]
    ):
        self._store_manager = StoreManager()
        for store_name, store in stores.items():
            if not isinstance(store, Store):
                raise TypeError(f"Expected Store, got {type(store)}")
            logger.debug(f"Adding store {store_name} to the database.")
            self.stores.add_store(store_name, store)

    @property
    def stores(self) -> StoreManager:
        """Returns the store manager."""
        return self._store_manager

    @property
    def metadata_stores(self) -> StoreManager:
        return StoreManager(
            self.stores.metadata_stores
        )

    @property
    def data_stores(self) -> StoreManager:
        """Alias for stores property."""
        return StoreManager(
            self.stores.data_stores
        )

    # @abstractmethod
    # def linked_upload(self, filename: Union[str, pathlib.Path]):
    #     """Uploads the file to both stores and links them."""

    # def execute_query(self, store_name: str, query: Query) -> QueryResult:
    #     return self.store_manager.execute_query(store_name, query)



import logging
import pathlib
import sys
import unittest
from typing import List

from gldb import GenericLinkedDatabase, DataStore, MetadataStore
from gldb.query import QueryResult, FederatedQueryResult, SparqlQuery
from gldb.stores import RDFStore

logger = logging.getLogger("gldb")
logger.setLevel(logging.DEBUG)
for h in logger.handlers:
    h.setLevel(logging.DEBUG)

__this_dir__ = pathlib.Path(__file__).parent

sys.path.insert(0, str(__this_dir__))
from gldb.stores import InMemoryRDFStore
from example_storage_db import CSVDatabase


def get_temperature_data_by_date(db, date: str) -> List[FederatedQueryResult]:
    """High-level abstraction for user to find temperature data.
    It is a federated query that combines metadata and data from the RDF and CSV databases."""
    sparql_query = """
    PREFIX dcterms: <http://purl.org/dc/terms/>
    PREFIX dcat: <http://www.w3.org/ns/dcat#>

    SELECT ?dataset ?url
    WHERE {{
      ?dataset a dcat:Dataset .
      ?dataset dcterms:created "{date}" .
      ?dataset dcat:distribution ?distribution .
      ?distribution dcat:downloadURL ?url .
    }}
    """.format(date=date)
    # results = self["rdf_database"].execute_query(SparqlQuery(sparql_query))
    _store: RDFStore = db.stores.rdf_database
    results = SparqlQuery(sparql_query).execute(_store)

    # result_data = [{str(k): parse_literal(v) for k, v in binding.items()} for binding in results.data.bindings]

    federated_query_results = []

    rdf_database = db.stores.rdf_database
    for dataset, url in zip(results.data["dataset"], results.data["url"]):
        filename = str(url).rsplit('/', 1)[-1]

        data = db.stores.csv_database.get_all(filename)

        # query all metadata for the dataset:
        metadata_sparql = """
        PREFIX dcterms: <http://purl.org/dc/terms/>
        PREFIX dcat: <http://www.w3.org/ns/dcat#>

        SELECT ?p ?o
        WHERE {{
          <{dataset}> ?p ?o .
        }}
        """.format(dataset=dataset)
        metadata = SparqlQuery(metadata_sparql).execute(rdf_database)  # .execute(rdf_database.graph)
        # dataset_result_data = [{str(k): v for k, v in binding.items()} for binding in
        #                        metadata_result.data.bindings]
        # metadata = {d["p"]: d["o"] for d in dataset_result_data}
        # context = {"dcterms": "http://purl.org/dc/terms/", "dcat": "http://www.w3.org/ns/dcat#",
        #            "ex": "https://example.org/"}
        #
        # g = rdflib.Graph()
        # for k, v in metadata.items():
        #     g.add((rdflib.URIRef(res["dataset"]), rdflib.URIRef(k), v))
        # jsonld = g.serialize(format="json-ld", context=context)
        # jsonld_dict = json.loads(jsonld)
        # for k, v in jsonld_dict.items():
        #     if isinstance(v, dict):
        #         if len(v) == 1 and "@id" in v:
        #             jsonld_dict[k] = v["@id"]
        federated_query_results.append(FederatedQueryResult(data=data, metadata=metadata))
        # better convert metadata to json-ld string

    return federated_query_results


class TestGenericLinkedDatabase(unittest.TestCase):

    def test_rdf_and_csv_stores(self):
        with self.assertRaises(TypeError):
            GenericLinkedDatabase(
                stores={
                    "rdf_database": 2,
                    "csv_database": "not_a_store"
                }
            )

        with self.assertRaises(ValueError):
            db = GenericLinkedDatabase(
                stores={
                    "rdf_database": InMemoryRDFStore(__this_dir__ / "data")
                }
            )
            db.stores.add_store("rdf_database", CSVDatabase())

        db = GenericLinkedDatabase(
            stores={
                "rdf_database": InMemoryRDFStore(__this_dir__ / "data"),
                "csv_database": CSVDatabase()
            }
        )

        rdf_database: RDFStore = db.stores.rdf_database
        csv_database: DataStore = db.stores.csv_database

        self.assertIsInstance(rdf_database, MetadataStore)
        self.assertIsInstance(rdf_database, InMemoryRDFStore)
        self.assertIsInstance(csv_database, DataStore)
        self.assertIsInstance(csv_database, CSVDatabase)

        csv_database = db.data_stores.csv_database
        rdf_database = db.metadata_stores.rdf_database
        self.assertIsInstance(csv_database, CSVDatabase)
        self.assertIsInstance(rdf_database, InMemoryRDFStore)

        rdf_database.upload_file(__this_dir__ / "data/data1.jsonld")

        query = SparqlQuery(query="SELECT * WHERE {?s ?p ?o}", description="Selects all triples")
        res = query.execute(rdf_database)
        self.assertEqual(res.description, "Selects all triples")

        self.assertIsInstance(res, QueryResult)
        print(res.data)
        # self.assertIn(25, sorted([i.get("foaf:age", -1) for i in res.data["@graph"]]))
        # self.assertIn(30, sorted([i.get("foaf:age", -1) for i in res.data["@graph"]]))

        rdf_database.upload_file(__this_dir__ / "data/metadata.jsonld")

        csv_database.upload_file(__this_dir__ / "data/random_data.csv")
        csv_database.upload_file(__this_dir__ / "data/random_data.csv")
        csv_database.upload_file(__this_dir__ / "data/temperature.csv")
        csv_database.upload_file(__this_dir__ / "data/users.csv")

        data = get_temperature_data_by_date(db, date="2024-01-01")
        self.assertIsInstance(data, list)
        self.assertIsInstance(data[0], FederatedQueryResult)



import json
from datetime import datetime, date, time
from decimal import Decimal

import pandas as pd
from rdflib import Graph

# --- minimal caster for common XSD datatypes ---
_XSD = "http://www.w3.org/2001/XMLSchema#"
_NUM_DT = {
    _XSD + "integer", _XSD + "int", _XSD + "long", _XSD + "short",
    _XSD + "byte", _XSD + "nonNegativeInteger", _XSD + "nonPositiveInteger",
    _XSD + "positiveInteger", _XSD + "negativeInteger"
}
_FLOAT_DT = {_XSD + "float", _XSD + "double"}
_DEC_DT = {_XSD + "decimal"}


def sparql_query_to_jsonld(graph: Graph, query: str) -> dict:
    results = graph.query(query)

    jsonld = {
        "@context": {},
        "@graph": []
    }

    for row in results:
        item = {}
        for var, value in row.asdict().items():
            if value is not None:
                if hasattr(value, 'n3'):
                    val_str = value.n3(graph.namespace_manager)
                else:
                    val_str = str(value)

                item[var] = val_str

                # Add simple context mapping
                if isinstance(value, (str, int, float)):
                    jsonld["@context"][var] = None
                elif value.__class__.__name__ == 'URIRef':
                    jsonld["@context"][var] = str(value)

        if item:
            jsonld["@graph"].append(item)

    return jsonld


def _cast_cell(cell, cast_literals: bool):
    """cell is a SPARQL JSON binding object like {'type': 'literal', 'value': '42', 'datatype': '...'}"""
    if cell is None:
        return None
    t = cell.get("type")
    v = cell.get("value")
    if not cast_literals:
        return v
    if t == "literal":
        dt = cell.get("datatype")
        if dt in _NUM_DT:
            try:
                return int(v)
            except Exception:
                return v
        if dt in _FLOAT_DT:
            try:
                return float(v)
            except Exception:
                return v
        if dt in _DEC_DT:
            try:
                return Decimal(v)
            except Exception:
                return v
        if dt == _XSD + "boolean":
            return v.lower() == "true"
        if dt == _XSD + "dateTime":
            try:
                return datetime.fromisoformat(v.replace("Z", "+00:00"))
            except Exception:
                return v
        if dt == _XSD + "date":
            try:
                return date.fromisoformat(v)
            except Exception:
                return v
        if dt == _XSD + "time":
            try:
                return time.fromisoformat(v)
            except Exception:
                return v
        # language-tagged literal like {"xml:lang": "en"}
        # fall through to string value
        return v
    # URIs / bnodes: just return string
    return v


def sparql_json_to_dataframe(results_json: dict, cast_literals: bool = True) -> pd.DataFrame:
    """
    Turn a SPARQL SELECT JSON result into a pandas DataFrame.
    Columns are exactly results['head']['vars'] and missing bindings become None.
    """
    vars_ = results_json.get("head", {}).get("vars", [])
    rows = []
    for b in results_json.get("results", {}).get("bindings", []):
        row = {var: _cast_cell(b.get(var), cast_literals) for var in vars_}
        rows.append(row)
    return pd.DataFrame(rows, columns=vars_)


if __name__ == "__main__":
    # Example usage
    g = Graph()
    g.parse(data="""
    @prefix ex: <http://example.org/> .
    ex:subject1 ex:predicate1 "object1" .
    ex:subject2 ex:predicate2 "object2" .
    """, format="turtle")

    query = """
    SELECT ?s ?p ?o WHERE {
        ?s ?p ?o .
    }
    """
    jsonld_result = sparql_query_to_jsonld(g, query)
    print(json.dumps(jsonld_result, indent=2))

    query2 = """
    PREFIX ex: <http://example.org/>
    SELECT ?p ?o WHERE {
        ex:subject1 ?p ?o .
    }
    """
    jsonld_result = sparql_query_to_jsonld(g, query2)
    print(json.dumps(jsonld_result, indent=2))



from .data_store_query import DataStoreQuery
from .metadata_query import MetadataStoreQuery
from .metadata_query import SparqlQuery, RemoteSparqlQuery
from .query import Query, QueryResult, FederatedQueryResult



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Dict
from typing import Optional

from rdflib import Graph

from ..stores import Store


class AbstractQuery(ABC):

    @abstractmethod
    def execute(self, store: Store) -> "QueryResult":
        """Executes the query."""


class QueryResult:

    def __init__(self, query: AbstractQuery, data: Any, description: Optional[str] = None,
                 derived_graph: Optional[Graph] = None):
        self.query = query
        self.data = data
        self.description = description
        self.derived_graph = derived_graph

    def __len__(self):
        return len(self.data)

    def __repr__(self):
        return f"{self.__class__.__name__}(\n  query={self.query},\n  data={self.data},\n  description={self.description}\n)"


@dataclass(frozen=True)
class FederatedQueryResult:
    data: Any
    metadata: Dict


class Query(AbstractQuery, ABC):

    def __init__(self, query, description=None):
        self.query = query
        self.description = description

    def __eq__(self, other):
        if not isinstance(other, Query):
            return NotImplemented
        return self.query == other.query and self.description == other.description

    def __repr__(self):
        description_repr = self.description or ""
        return f"{self.__class__.__name__}(query=\"{self.query}\", description=\"{description_repr}\")"



from abc import ABC

import pandas as pd
import rdflib

from gldb import logger
from gldb.query.query import Query, QueryResult
from gldb.stores import RDFStore, RemoteSparqlStore
from .utils import sparql_json_to_dataframe


def parse_literal(literal):
    if isinstance(literal, rdflib.Literal):
        return literal.value
    if isinstance(literal, rdflib.URIRef):
        return str(literal)
    return literal


def sparql_result_to_df(bindings):
    return pd.DataFrame([{str(k): parse_literal(v) for k, v in binding.items()} for binding in bindings])


class MetadataStoreQuery(Query, ABC):
    """RDF Store Query interface."""


class SparqlQuery(MetadataStoreQuery):
    """A SPARQL query interface for RDF stores."""

    def execute(self, store: RDFStore, *args, **kwargs) -> QueryResult:
        """Execute the SPARQL query against the given RDF store.

        Parameters
        ----------
        store : RDFStore
            The RDF store to execute the query against.
        *args
            Additional positional arguments to pass to the store's query method.
        **kwargs
            Additional keyword arguments to pass to the store's query method.

        Returns
        -------
        QueryResult
            The result of the query execution.
        """
        if isinstance(store, RemoteSparqlStore):
            return RemoteSparqlQuery(self.query, self.description).execute(store)
        res = store.graph.query(self.query, *args, **kwargs)
        bindings = res.bindings
        try:
            derived_graph = res.graph
        except AttributeError:
            derived_graph = None
        if bindings is None:
            return QueryResult(
                query=self,
                data=pd.DataFrame(),
                description=self.description,
                derived_graph=derived_graph
            )
        return QueryResult(
            query=self,
            data=sparql_result_to_df(bindings),
            description=self.description,
            derived_graph=derived_graph
        )


class RemoteSparqlQuery(MetadataStoreQuery):

    def execute(self, store: RemoteSparqlStore, *args, **kwargs) -> QueryResult:
        sparql = store.wrapper
        sparql.setQuery(self.query)

        results = sparql.queryAndConvert()

        try:
            data = sparql_json_to_dataframe(results)
        except Exception as e:
            logger.debug("Failed to convert SPARQL results to DataFrame: %s", e)
            data = results
        return QueryResult(
            query=self,
            data=data,
            description=self.description
        )



import sys
import unittest

from gldb.query import Query, QueryResult, RemoteSparqlQuery
from gldb.stores import DataStore
from gldb.stores import RemoteSparqlStore
from gldb.stores import StoreManager


class MockSqlQuery(Query):

    def __init__(self, query: str, description: str = None):
        # super().__init__(query, description)
        self.query = query

    def execute(self, *args, **kwargs) -> QueryResult:
        return QueryResult(self, "mock_result")


class CSVDatabase(DataStore):

    def __init__(self):
        self._filenames = []
        self.tables = {}
        self._expected_file_extensions = {".csv", }

    # @property
    # def query(self) -> Type[Query]:
    #     return MockSqlQuery

    def upload_file(self, filename) -> bool:
        return True

    def execute_query(self, query: Query):
        raise NotImplementedError("CSVDatabase does not support queries.")


class TestDataStore(unittest.TestCase):

    def test_DataStoreManager(self):
        manager = StoreManager()
        self.assertEqual(len(manager), 0)
        self.assertEqual(
            manager.__repr__(),
            "StoreManager(stores=[])"
        )

    def test_add_store(self):
        manager = StoreManager()
        store = CSVDatabase()
        self.assertEqual(
            len(manager.stores),
            0
        )
        self.assertEqual(store.__repr__(), "CSVDatabase()")
        manager.add_store("test_store", store)

        self.assertEqual(len(manager), 1)

        with self.assertRaises(AttributeError):
            manager.does_not_exist

    def test_query_store(self):
        store = CSVDatabase()
        query = MockSqlQuery(query="SELECT * FROM test_table;")
        self.assertIsInstance(query, Query)
        self.assertIsInstance(query, MockSqlQuery)
        self.assertEqual(query.query, "SELECT * FROM test_table;")

        qres = query.execute(store)
        self.assertIsInstance(qres, QueryResult)
        self.assertEqual(qres.data, "mock_result")

    def test_wikidata_store(self):
        if not (sys.version_info.major == 3 and sys.version_info.minor == 12):
            self.skipTest("Skipping test on non-3.12 Python to avoid rate limiting")
        remote_store = RemoteSparqlStore("https://query.wikidata.org/sparql", return_format="json")
        self.assertIsInstance(remote_store, RemoteSparqlStore)

        sparql_query = """
SELECT * WHERE {
  wd:Q131549102 ?property ?value.
  OPTIONAL { ?value rdfs:label ?valueLabel. }
}
ORDER BY ?propertyLabel
"""
        query = RemoteSparqlQuery(sparql_query)
        res = query.execute(remote_store)
        self.assertTrue(len(res.data) >= 172)



import pathlib
import sys
import unittest

import rdflib
from SPARQLWrapper import SPARQLWrapper, JSON

from gldb.query import Query, QueryResult, SparqlQuery, RemoteSparqlQuery
from gldb.query.metadata_query import sparql_result_to_df
from gldb.stores import InMemoryRDFStore, RemoteSparqlStore

__this_dir__ = pathlib.Path(__file__).parent.resolve()

TESTING_VERSIONS = (9, 12)


def get_python_version():
    """Get the current Python version as a tuple."""
    return sys.version_info.major, sys.version_info.minor, sys.version_info.micro


class TestQuery(unittest.TestCase):

    def test_query(self):
        class SQLQuery(Query):

            def execute(self) -> QueryResult:
                return QueryResult(self, data="result", description=self.description)

        q = SQLQuery("SELECT * FROM Customers;", "Get all customers")
        res = q.execute()
        self.assertEqual(res.query, q)
        self.assertEqual(res.description, "Get all customers")

        res = SQLQuery("SELECT * FROM Customers;", "Get all customers").execute()
        self.assertIsInstance(res, QueryResult)

    def test_sparql_query(self):
        graph = rdflib.Graph()
        sparql_query = SparqlQuery(query="SELECT * WHERE { ?s ?p ?o }")
        self.assertEqual(
            sparql_query.__repr__(),
            'SparqlQuery(query="SELECT * WHERE { ?s ?p ?o }", description="")'
        )
        store = InMemoryRDFStore(data_dir=__this_dir__ / "testdata")
        res = SparqlQuery("SELECT * WHERE { ?s ?p ?o }").execute(store)
        self.assertIsInstance(res, QueryResult)
        self.assertEqual(res.query, sparql_query)
        self.assertEqual(res.derived_graph, None)
        self.assertTrue(res.data.equals(sparql_result_to_df(graph.query("SELECT * WHERE { ?s ?p ?o }"))))

    @unittest.skipUnless(get_python_version()[1] in TESTING_VERSIONS,
                         reason="Nur auf Python 3.9 und 3.12 testen")
    def test_wikidata_query(self):
        enpoint_url = "https://query.wikidata.org/sparql"
        sparql_wrapper = SPARQLWrapper(enpoint_url)
        sparql_wrapper.setReturnFormat(JSON)
        query_str = """
SELECT * WHERE {
  wd:Q131448345 ?property ?value.
  OPTIONAL { ?value rdfs:label ?valueLabel. }
}
ORDER BY ?propertyLabel
"""
        sparql_query = RemoteSparqlQuery(query_str)
        self.assertEqual(
            sparql_query.__repr__(),
            f"RemoteSparqlQuery(query=\"{query_str}\", description=\"\")"
        )
        remote_store = RemoteSparqlStore(endpoint_url=enpoint_url, return_format="json")

        res = sparql_query.execute(remote_store)
        self.assertIsInstance(res, QueryResult)
        self.assertEqual(res.query, sparql_query)
        self.assertTrue(len(res.data) >= 917)

        sparql_query = RemoteSparqlQuery(query="DESCRIBE wd:Q131448345")
        remote_store = RemoteSparqlStore(endpoint_url=enpoint_url, return_format="json-ld")
        res = sparql_query.execute(remote_store)
        print(res.data.serialize("json-ld").serialize())

    def test_construct_query(self):
        # g = rdflib.Graph()
        # g.parse(__this_dir__ / "data" / "planets.ttl", format="turtle")
        construct_query = """
        PREFIX ex: <http://example.org/schema/>
        PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>

        CONSTRUCT {
          ?planet a ex:Planet ;
                  ex:mass ?mass ;
                  ex:radius ?radius ;
                  ex:label ?label .
        }
        WHERE {
          ?planet a ex:Planet ;
                  ex:mass ?mass ;
                  ex:radius ?radius .
          OPTIONAL {
            ?planet rdfs:label ?label .
            FILTER(LANG(?label) = "en")
          }
        }
        """

        sparql_query = SparqlQuery(query=construct_query)
        store = InMemoryRDFStore(data_dir=__this_dir__ / "data")
        res = sparql_query.execute(store)

        self.assertEqual(16, len(res.derived_graph))

        select_query = """
        PREFIX ex: <http://example.org/schema/>

        SELECT ?planet ?mass ?radius ?label
        WHERE {
          ?planet a ex:Planet ;
                  ex:mass ?mass ;
                  ex:radius ?radius .
          OPTIONAL { ?planet ex:label ?label }

          FILTER(?mass > 1.0e24)
        }
        ORDER BY DESC(?mass)
        """
        self.assertEqual(3, len(res.derived_graph.query(select_query)))



from unittest.mock import MagicMock
from unittest.mock import patch, Mock

import pandas as pd
import pytest

from gldb.query.metadata_query import RemoteSparqlQuery
from gldb.stores import GraphDB


def make_graphdb():
    return GraphDB(endpoint="http://localhost:7200",
                   repository="testrepo",
                   username="user",
                   password="pass")


@patch.object(GraphDB, "get_repository_info", return_value={})
@patch("requests.post")
def test_upload_file_success(mock_post, mock_repo_info, tmp_path):
    # Erfolgreicher Upload
    test_file = tmp_path / "test.ttl"
    test_file.write_text("@prefix ex: <http://example.org/> . ex:a ex:b ex:c .")
    mock_post.return_value = Mock(status_code=204)
    db = make_graphdb()
    assert db.upload_file(test_file) is True
    mock_post.assert_called_once()


@patch.object(GraphDB, "get_repository_info", return_value={})
@patch("requests.post")
def test_upload_file_not_found(mock_post, mock_repo_info):
    db = make_graphdb()
    with pytest.raises(FileNotFoundError):
        db.upload_file("notfound.ttl")
    mock_post.assert_not_called()


@patch.object(GraphDB, "get_repository_info", return_value={})
@patch("requests.post")
def test_upload_file_wrong_format(mock_post, mock_repo_info, tmp_path):
    test_file = tmp_path / "test.txt"
    test_file.write_text("dummy")
    db = make_graphdb()
    with pytest.raises(ValueError):
        db.upload_file(test_file)
    mock_post.assert_not_called()


@patch.object(GraphDB, "get_repository_info", return_value={})
@patch("requests.post")
def test_upload_file_server_error(mock_post, mock_repo_info, tmp_path):
    test_file = tmp_path / "test.ttl"
    test_file.write_text("@prefix ex: <http://example.org/> . ex:a ex:b ex:c .")
    mock_post.return_value = Mock(status_code=500, text="Internal Server Error")
    db = make_graphdb()
    with pytest.raises(RuntimeError):
        db.upload_file(test_file)
    mock_post.assert_called_once()


@patch.object(GraphDB, "get_repository_info", return_value={})
def test_select_all(mock_repo_info):
    # SELECT_ALL als RemoteSparqlQuery
    SELECT_ALL = RemoteSparqlQuery(
        "SELECT * WHERE { ?s ?p ?o }",
        description="Selects all triples in the RDF database"
    )
    db = make_graphdb()

    # Replace the wrapper on the instance with a mock
    fake = MagicMock()
    fake.queryAndConvert.return_value = {
        "head": {"vars": ["s", "p", "o"]},
        "results": {
            "bindings": [
                {"s": {"value": "s1"}, "p": {"value": "p1"}, "o": {"value": "o1"}},
                {"s": {"value": "s2"}, "p": {"value": "p2"}, "o": {"value": "o2"}},
            ]
        },
    }
    db._wrapper = fake

    result = SELECT_ALL.execute(db)
    pd.testing.assert_frame_equal(
        result.data,
        pd.DataFrame(
            {
                "s": ["s1", "s2"],
                "p": ["p1", "p2"],
                "o": ["o1", "o2"],
            }
        )
    )
    # assert result.data["results"]["bindings"] == [
    #     {"s": {"value": "s1"}, "p": {"value": "p1"}, "o": {"value": "o1"}},
    #     {"s": {"value": "s2"}, "p": {"value": "p2"}, "o": {"value": "o2"}},
    # ]
    fake.setQuery.assert_called_once_with("SELECT * WHERE { ?s ?p ?o }")
    fake.queryAndConvert.assert_called_once()



import pathlib
from typing import Type

import pandas as pd

from gldb import DataStore
from gldb.query import Query


class CSVDbQuery(Query):
    def __init__(self, query: str, description: str = None):
        super().__init__(query, description)
        self.query = query

    def execute(self, query, description=None, *args, **kwargs):
        raise NotImplementedError("CSVDbQuery does not support execution.")


class CSVDatabase(DataStore):

    def __init__(self):
        self._filenames = []
        self.tables = {}
        self._expected_file_extensions = {".csv", }

    @property
    def query(self) -> Type[CSVDbQuery]:
        return CSVDbQuery

    @property
    def expected_file_extensions(self):
        return self._expected_file_extensions

    def upload_file(self, filename) -> bool:
        filename = pathlib.Path(filename)
        assert filename.exists(), f"File {filename} does not exist."
        assert filename.suffix in self._expected_file_extensions, f"File type {filename.suffix} not supported"
        if filename.resolve().absolute() in self._filenames:
            return True
        self._filenames.append(filename.resolve().absolute())

        table_name = filename.name

        self.tables[table_name] = pd.read_csv(filename)
        return True

    def execute_query(self, query: Query):
        raise NotImplementedError("CSVDatabase does not support queries.")

    def get_all(self, table_name):
        return self.tables[table_name]



import pathlib
import shutil
from abc import ABC, abstractmethod
from typing import Dict, Union, Any

import rdflib
import requests

from gldb import logger


class Store(ABC):
    """Store interface."""

    # @property
    # @abstractmethod
    # def query(self) -> Type[Query]:
    #     """Returns the query class for the store."""

    @abstractmethod
    def upload_file(self, filename: Union[str, pathlib.Path]) -> Any:
        """Uploads a file to the store."""

    def __repr__(self):
        """String representation of the Store."""
        return f"{self.__class__.__name__}()"


class DataStore(Store, ABC):
    """Data store interface (concrete implementations can be sql or non sql databases)."""

    @abstractmethod
    def upload_file(self, filename: Union[str, pathlib.Path]):
        """Insert data into the data store."""


class MetadataStore(Store, ABC):
    """Metadata database interface using."""

    @abstractmethod
    def upload_file(self, filename: Union[str, pathlib.Path]) -> bool:
        """Insert data into the data store."""


class RDFStore(MetadataStore, ABC):
    namespaces = {
        "ex": "https://example.org/",
        "afn": "http://jena.apache.org/ARQ/function#",
        "agg": "http://jena.apache.org/ARQ/function/aggregate#",
        "apf": "http://jena.apache.org/ARQ/property#",
        "array": "http://www.w3.org/2005/xpath-functions/array",
        "dcat": "http://www.w3.org/ns/dcat#",
        "dcterms": "http://purl.org/dc/terms/",
        "fn": "http://www.w3.org/2005/xpath-functions",
        "foaf": "http://xmlns.com/foaf/0.1/",
        "geoext": "http://rdf.useekm.com/ext#",
        "geof": "http://www.opengis.net/def/function/geosparql/",
        "gn": "http://www.geonames.org/ontology#",
        "graphdb": "http://www.ontotext.com/config/graphdb#",
        "list": "http://jena.apache.org/ARQ/list#",
        "local": "https://doi.org/10.5281/zenodo.14175299/",
        "m4i": "http://w3id.org/nfdi4ing/metadata4ing#",
        "map": "http://www.w3.org/2005/xpath-functions/map",
        "math": "http://www.w3.org/2005/xpath-functions/math",
        "ofn": "http://www.ontotext.com/sparql/functions/",
        "omgeo": "http://www.ontotext.com/owlim/geo#",
        "owl": "http://www.w3.org/2002/07/owl#",
        "path": "http://www.ontotext.com/path#",
        "prov": "http://www.w3.org/ns/prov#",
        "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
        "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
        "rep": "http://www.openrdf.org/config/repository#",
        "sail": "http://www.openrdf.org/config/sail#",
        "schema": "https://schema.org/",
        "spif": "http://spinrdf.org/spif#",
        "sr": "http://www.openrdf.org/config/repository/sail#",
        "ssno": "https://matthiasprobst.github.io/ssno#",
        "wgs": "http://www.w3.org/2003/01/geo/wgs84_pos#",
        "xsd": "http://www.w3.org/2001/XMLSchema#"
    }

    @property
    @abstractmethod
    def graph(self) -> rdflib.Graph:
        """Return graph for the metadata store."""

    # @property
    # def query(self):
    #     return SparqlQuery(self.graph)


class RemoteSparqlStore(MetadataStore):

    def __init__(self, endpoint_url, return_format: str = None):
        try:
            from SPARQLWrapper import SPARQLWrapper
        except ImportError:
            raise ImportError("Please install SPARQLWrapper to use this class: 'pip install SPARQLWrapper'")

        self._wrapper = SPARQLWrapper(endpoint_url)
        if return_format is not None:
            self._wrapper.setReturnFormat(return_format)

    @property
    def wrapper(self):
        return self._wrapper

    # @property
    # def query(self) -> RemoteSparqlQuery:
    #     """Return graph for the DbPedia metadata store."""
    #     try:
    #         from SPARQLWrapper import SPARQLWrapper, JSON
    #     except ImportError:
    #         raise ImportError("Please install SPARQLWrapper to use this class: pip install SPARQLWrapper")
    #     sparql = SPARQLWrapper(self.endpoint)
    #     sparql.setReturnFormat(JSON)
    #     return RemoteSparqlQuery(sparql)

    def upload_file(self, filename: Union[str, pathlib.Path]) -> bool:
        """Uploads a file to the remote SPARQL endpoint."""
        raise NotImplementedError("Remote SPARQL Store does not support file uploads.")


class StoreManager:
    """Store manager that manages the interaction between stores."""

    def __init__(self, stores: Dict[str, Store] = None):
        self._stores: Dict[str, DataStore] = stores if stores is not None else {}

    def __getattr__(self, item) -> Store:
        """Allows access to stores as attributes."""
        if item in self.stores:
            return self.stores[item]
        return super().__getattribute__(item)

    def __len__(self):
        """Returns the number of stores managed."""
        return len(self.stores)

    def __repr__(self):
        """String representation of the DataStoreManager."""
        store_names = ", ".join(self.stores.keys())
        return f"{self.__class__.__name__}(stores=[{store_names}])"

    @property
    def stores(self) -> Dict[str, Store]:
        """Returns the stores managed by the manager."""
        return self._stores

    @property
    def data_stores(self) -> Dict[str, DataStore]:
        """Alias for stores property."""
        return {k: v for k, v in self.stores.items() if isinstance(v, DataStore)}

    @property
    def metadata_stores(self) -> Dict[str, MetadataStore]:
        """Alias for stores property."""
        return {k: v for k, v in self.stores.items() if isinstance(v, MetadataStore)}

    def add_store(self, store_name: str, store: Store):
        """Add a new store to the manager."""
        if store_name in self.stores:
            raise ValueError(f"DataStore with name {store_name} already exists.")
        self.stores[store_name] = store


# concrete implementations of Store

class InMemoryRDFStore(RDFStore):
    """In-memory RDF database that can upload files and return a combined graph."""

    _expected_file_extensions = {".ttl", ".rdf", ".jsonld", ".nt", ".xml", ".n3"}

    def __init__(
            self,
            data_dir: Union[str, pathlib.Path],
            recursive_exploration: bool = False,
            formats=None
    ):
        if formats is None:
            formats = self._expected_file_extensions
        elif isinstance(formats, str):
            formats = {f".{formats.lstrip('.')}", }
        elif isinstance(formats, (list, set, tuple)):
            formats = {f".{fmt.lstrip('.')}" for fmt in formats}
        else:
            raise ValueError("formats must be a string or a list/set/tuple of strings.")

        for _fmt in formats:
            if _fmt not in self._expected_file_extensions:
                raise ValueError(f"File format '{_fmt}' not supported.")

        self._expected_file_extensions = formats
        self._data_dir = pathlib.Path(data_dir).resolve()
        self._data_dir.mkdir(parents=True, exist_ok=True)
        self._recursive_exploration = recursive_exploration
        self._filenames = []
        self._graphs = {}
        self._combined_graph = rdflib.Graph()
        self.update()

    @property
    def data_dir(self) -> pathlib.Path:
        """Returns the data directory where files are stored."""
        return self._data_dir

    def update(self):
        for _ext in self._expected_file_extensions:
            if self._recursive_exploration:
                self._filenames.extend([f.resolve().absolute() for f in self.data_dir.rglob(f"*{_ext}")])
            else:
                self._filenames.extend([f.resolve().absolute() for f in self.data_dir.glob(f"*{_ext}")])
        self._filenames = list(set(self._filenames))  # remove duplicates
        for filename in self._filenames:
            self._add_to_graph(filename)

    @property
    def filenames(self):
        """Returns the list of filenames uploaded to the store."""
        return self._filenames

    def upload_file(self, filename) -> bool:
        filename = pathlib.Path(filename).resolve().absolute()
        if not filename.exists():
            raise FileNotFoundError(f"File {filename} not found.")
        if filename.suffix not in self._expected_file_extensions:
            raise ValueError(f"File type {filename.suffix} not supported.")
        if filename in self._filenames:
            self._filenames.remove(filename)
        self._filenames.append(filename)
        if filename.parent != self.data_dir:
            shutil.copy(filename, self.data_dir / filename.name)
        self._add_to_graph(filename)
        return True

    def _add_to_graph(self, filename: pathlib.Path):
        """Adds the RDF graph from the file to the combined graph."""
        g = self._graphs.get(filename, None)
        if not g:
            g = rdflib.Graph()
            try:
                g.parse(filename)
            except Exception as e:
                raise ValueError(f"Could not parse file '{filename}'. Error: {e}")
            for s, p, o in g:
                if isinstance(s, rdflib.BNode):
                    new_s = rdflib.URIRef(f"https://example.org/{s}")
                else:
                    new_s = s
                if isinstance(o, rdflib.BNode):
                    new_o = rdflib.URIRef(f"https://example.org/{o}")
                else:
                    new_o = o
                g.remove((s, p, o))
                g.add((new_s, p, new_o))
            self._graphs[filename] = g
            self._combined_graph += g

    @property
    def graph(self) -> rdflib.Graph:
        return self._combined_graph


class GraphDB(RemoteSparqlStore):
    """GraphDB RDF database store."""

    def __init__(self, endpoint: str, repository: str, username: str = None, password: str = None):
        super().__init__(f"{endpoint}/repositories/{repository}")
        try:
            from SPARQLWrapper import SPARQLWrapper, JSON
        except ImportError:
            raise ImportError("Please install SPARQLWrapper to use this class: pip install SPARQLWrapper")
        self._endpoint = endpoint
        self._repository = repository
        self._username = username
        self._password = password
        self._wrapper.setReturnFormat(JSON)
        if self.username and self.password:
            self._wrapper.setCredentials(self.username, self.password)

        repo_info = self.get_repository_info(self.repository)
        if not repo_info:
            logger.info(f"The repository '{self.repository}' does not exist. "
                        "Call create_repository('config.ttl') with a valid configuration "
                        "('config.ttl') file to create it.")

    @property
    def endpoint(self) -> str:
        return self._endpoint

    @property
    def repository(self) -> str:
        return self._repository

    @property
    def username(self) -> str:
        return self._username

    @property
    def password(self) -> str:
        return self._password

    def upload_file(self, filename: Union[str, pathlib.Path]) -> bool:
        """Uploads an RDF file to das GraphDB-Repository."""
        filename = pathlib.Path(filename).resolve().absolute()
        if not filename.exists():
            raise FileNotFoundError(f"File {filename} not found.")
        # Determine content type based on file extension
        ext = filename.suffix.lower()
        if ext == ".ttl":
            content_type = "text/turtle"
        elif ext == ".rdf":
            content_type = "application/rdf+xml"
        elif ext == ".jsonld":
            content_type = "application/ld+json"
        else:
            raise ValueError(f"File form '{ext}' not supported.")
        # Target URL
        url = f"{self.endpoint}/repositories/{self.repository}/statements"
        with open(filename, "rb") as f:
            data = f.read()
        headers = {"Content-Type": content_type}
        auth = (self.username, self.password) if self.username and self.password else None
        response = requests.post(url, data=data, headers=headers, auth=auth)
        if response.status_code in (200, 201, 204):
            return True
        else:
            raise RuntimeError(f"Upload failed: {response.status_code} {response.text}")

    def create_repository(self, config_path: Union[str, pathlib.Path]) -> bool:
        """Creates a GraphDB repository using a configuration file (repo-config.ttl)."""
        config_path = pathlib.Path(config_path).resolve().absolute()
        if not config_path.exists():
            raise FileNotFoundError(f"Config file not found: {config_path}.")
        url = f"{self.endpoint}/rest/repositories"
        auth = (self.username, self.password) if self.username and self.password else None
        with open(config_path, "rb") as f:
            files = {"config": (config_path.name, f, "application/x-turtle")}
            response = requests.post(url, files=files, auth=auth)
        if response.status_code in (201, 204):
            return True
        else:
            raise RuntimeError(f"Repository konnte nicht angelegt werden: {response.status_code} {response.text}")

    def list_repositories(self) -> Any:
        """Lists all repositories in the GraphDB instance."""
        url = f"{self.endpoint}/rest/repositories"
        headers = {"Accept": "application/json"}
        auth = (self.username, self.password) if self.username and self.password else None
        response = requests.get(url, headers=headers, auth=auth)
        if response.status_code == 200:
            return response.json()
        return None

    def get_repository_info(self, repository=None) -> Dict:
        """Returns information about a specific repository."""
        repo = repository or self.repository
        url = f"{self.endpoint}/rest/repositories/{repo}"
        headers = {"Accept": "application/json"}
        auth = (self.username, self.password) if self.username and self.password else None
        response = requests.get(url, headers=headers, auth=auth)

        if response.status_code == 200:
            return response.json()
        if response.status_code == 404:
            logger.debug(f"Repository '{repo}' does not exist.")
            return {}
        if response.status_code == 401:
            raise RuntimeError(f"Unauthorized access to repository '{repo}'. Check your credentials.")
        if response.status_code == 403:
            raise RuntimeError(f"Forbidden access to repository '{repo}'. Check your permissions.")
        raise RuntimeError(f"Error fetching repository info: {response.status_code} {response.text}")

    def get_or_create_repository(self, config_path: Union[str, pathlib.Path]) -> bool:
        """Gets the repository info if it exists, otherwise creates it using the provided config file."""
        if self.get_repository_info(self.repository):
            logger.debug(f"Found existing repository '{self.repository}'.")
            return True
        logger.debug(f"Repository '{self.repository}' does not exist. Creating it.")
        return self.create_repository(config_path)

    def count_triples(self, key: str = "total", repository: str = None) -> int:
        """Counts the number of triples in a repository.

        Parameters
        ----------
        key : str, optional
            The key to extract the count from the response. Default is "total". Other options are
            "explicit", "inferred".
        repository : str, optional
            The name of the repository to count triples in. If None, uses the default repository.
        """
        repo = repository or self.repository
        url = f"{self.endpoint}/rest/repositories/{repo}/size"
        auth = (self.username, self.password) if self.username and self.password else None
        response = requests.get(url, auth=auth)
        if response.status_code == 200:
            try:
                data = response.json()
                return int(data.get(key, 0))
            except Exception:
                raise RuntimeError(f"Error parsing the triple number: {response.text}")
        else:
            raise RuntimeError(f"Error fetching the triple number: {response.status_code} {response.text}")

    def restart_repository(self, repository: str = None) -> bool:
        """Restarts a repository in the GraphDB instance."""
        repo = repository or self.repository
        url = f"{self.endpoint}/rest/repositories/{repo}/restart"
        auth = (self.username, self.password) if self.username and self.password else None
        response = requests.post(url, auth=auth)
        if response.status_code in (200, 204):
            return True
        else:
            raise RuntimeError(f"Error restarting the repository: {response.status_code} {response.text}")

    def delete_repository(self, repository: str = None) -> bool:
        """Deletes a repository from the GraphDB instance."""
        repo = repository or self.repository
        url = f"{self.endpoint}/rest/repositories/{repo}"
        auth = (self.username, self.password) if self.username and self.password else None
        response = requests.delete(url, auth=auth)
        if response.status_code in (200, 204):
            return True
        else:
            raise RuntimeError(f"Error deleting the repository: {response.status_code} {response.text}")



import json
import pathlib
import unittest
from datetime import datetime

from gldb import __version__


def get_package_meta():
    """Reads codemeta.json and returns it as dict"""
    with open(__this_dir__ / '../codemeta.json', 'r') as f:
        codemeta = json.loads(f.read())
    return codemeta


__this_dir__ = pathlib.Path(__file__).parent


class TestVersion(unittest.TestCase):

    def test_version(self):
        this_version = 'x.x.x'
        setupcfg_filename = __this_dir__ / '../setup.cfg'
        with open(setupcfg_filename, 'r') as f:
            lines = f.readlines()
            for line in lines:
                if 'version' in line:
                    this_version = line.split(' = ')[-1].strip()
        self.assertEqual(__version__, this_version)

    def test_codemeta(self):
        """checking if the version in codemeta.json is the same as the one of the toolbox"""

        codemeta = get_package_meta()
        codemeta_version = codemeta["version"]
        if "-rc." in codemeta_version:
            codemeta_version = codemeta_version.replace("-rc.", "rc")
        self.assertEqual(__version__, codemeta_version)


    def test_citation_cff(self):
        """checking if the version in CITATION.cff is the same as the one of the ssnolib"""
        this_version = 'x.x.x'
        dt = None
        setupcfg_filename = __this_dir__ / '../CITATION.cff'
        with open(setupcfg_filename, 'r') as f:
            lines = f.readlines()
            for line in lines:
                if 'version:' in line:
                    this_version = line.split(':')[-1].strip()
                elif 'date-released:' in line:
                    # check if the date is the same as the one in codemeta.json
                    date_str = line.split(':')[-1].strip()
                    dt = datetime.strptime(date_str, '%Y-%m-%d')
        if "-rc." in this_version:
            this_version = this_version.replace("-rc.", "rc")
        self.assertEqual(__version__, this_version)
        self.assertEqual(dt.strftime('%Y-%m-%d'), datetime.now().date().strftime('%Y-%m-%d'))